#!/usr/bin/perl -w
#
#  PRISM Object
#  Use this module to access PRISM files
#
#  Copyright (c) 2010  Willem A. Schreuder <willem@prinmath.com>
#  This file is part of the KUDU suite of software.
#
#  The KUDU programs are free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  The KUDU programs are distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along 
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
#  See also www.fsf.org

package PRISM;

use strict;
use FileHandle;
use IO::Seekable;

my %conv = (FT=>0.01/304.8 , IN=>0.01/25.4 , MM=>0.01);

#
#  Select value by interpolation
#
sub interpolate
{
   my ($f,$x0,$x1) = @_;
   return (defined($x0) && defined($x1)) ? (1-$f)*$x0 + $f*$x1 : (defined($x0) ? $x0 : $x1);
}

#
#  Create a new PRISM object
#
sub new
{
   my $class = shift;
   my $self = {};
   bless($self,$class);

   my ($type,$unit) = @_;
   $type = uc($type);
   $unit = uc($unit);
   ($type eq 'PPT') || die "Unknown PRISM type $type\n";
   exists($conv{$unit}) || die "Invalid PRISM units $unit\n";

   $self->{Ni} = 1405;
   $self->{Nj} =  621;
   $self->{X0} = -125.0;
   $self->{Y0} = 24.083333333333;
   $self->{X1} = $self->{X0} + $self->{Ni}/24;
   $self->{Y1} = $self->{Y0} + $self->{Nj}/24;
   $self->{TYPE} = $type;
   $self->{EXT}  = lc $type;
   $self->{CONV} = $conv{$unit};
   $self->{YR}   = -1;
   $self->{FD}   = undef;
   return $self;
}

#
#  Load PRISM data from file
#
sub load
{
   my $self = shift;
   my ($mo,$i,$j) = @_;
   defined($self->{FD}) || return;
   my $l = 3*($self->{Ni}*$self->{Nj}*$mo + $self->{Ni}*$j + $i);
   seek($self->{FD},$l,IO::Seekable::SEEK_SET) || die "Cannot seek PRISM $l\n";
   my $buf;
   (read($self->{FD},$buf,3)==3) || return undef;
   my $k = unpack('V',$buf.chr(0));
   ($k>0) && $k--;
   return $k;
}

#
#  Get PRISM data
#     $val = $prism->eval($lon,$lat,$yr,$mo);
#
sub eval
{
   my $self = shift;
   my ($x,$y,$yr,$mo) = @_;

   #  Check month
   defined($mo) || ($mo=0);
   ($mo<0 || $mo>12)  && die "Invalid PRISM month $mo\n";

   #  Trap coordinates out of range
   ($x<$self->{X0} || $x>$self->{X1} || $y<$self->{Y0} || $y>$self->{Y1}) && return;

   #  Open file for this year
   if ($yr != $self->{YR})
   {
      defined($self->{FD}) && $self->{FD}->close();
      $self->{FILE} = PM::grflib("PRISM/$self->{TYPE}/$yr.$self->{EXT}");
      if (-e $self->{FILE})
      {
         $self->{YR} = $yr;
         $self->{FD} = new FileHandle($self->{FILE});
         defined($self->{FD}) || die "Cannot open PRISM file $self->{FILE}\n";
         binmode($self->{FD});
      }
      else
      {
         $self->{FD} = undef;
      }
   }
   #  Get offset
   my $f = 24*($x-$self->{X0});
   my $g = 24*($y-$self->{Y0});
   my $i = int($f);
   my $j = int($g);
   ($i==$self->{Ni}) && $i--;
   ($j==$self->{Nj}) && $j--;
   $f -= $i;
   $g -= $j;
   my $z00 = $self->load($mo,$i  ,$j);
   my $z10 = $self->load($mo,$i+1,$j);
   my $z01 = $self->load($mo,$i  ,$j+1);
   my $z11 = $self->load($mo,$i+1,$j+1);

   my $in = interpolate($g , interpolate($f,$z00,$z10) , interpolate($f,$z01,$z11));
   return defined($in) ? $self->{CONV}*$in : undef;
}

#
#  Destructor
#
sub DESTROY
{
   my $self = shift;
   defined($self->{FD}) && $self->{FD}->close();
}
